import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
import scipy.spatial as spa

# constants
R = 0.5  # ball radius
P = np.load("small_p_filt4.npy")  # point cloud map
kd_tree = spa.KDTree(P.T)  # KD tree used to check for collision points

XDIM = [1, 23]  # allowed dimensions for planning
YDIM = [-3, 24]
ZDIM = [0, 8]

XDIM_A = [1, 20]  # allowed dimensions for planning - our hallway part
YDIM_A = [-3, 2]
ZDIM_A = [0, 8]
XDIM_B = [14, 23]  # allowed dimensions for planning - perpendicular hallway
YDIM_B = [-3, 24]
ZDIM_B = [0, 8]


EPSILON = 1.0    # step by which the ball can move in space
NUMNODES = 3000  # maximum number of nodes
GOAL_DISTANCE_THRESHOLD = 1.5

# Class representing a node
class Node:
    pos = np.array([[0, 0, 0]]).T
    parent = None

    def __init__(self, x, y, z):
        self.pos = np.array([[x,y,z]]).T

# Plots points of a pointcloud P onto axes ax
def showPointClouds(P,ax):
    ax.scatter(P[0, ::30], P[1, ::30], P[2, ::30], c='r', s=5, marker='.')

# Draws the tree graph stored in nodes list
def plotNodes(nodes,ax):
    # plot the whole tree
    for node in reversed(nodes):
        if node.parent is None:
            continue;
        ax.plot3D([node.parent.pos[0,0], node.pos[0,0]],
                  [node.parent.pos[1,0], node.pos[1,0]],
                  [node.parent.pos[2,0], node.pos[2,0]], color='b')
    # plot the winning trajectory
    node = nodes[-1]
    while not node.parent is None:
        ax.plot3D([node.parent.pos[0,0], node.pos[0,0]],
                  [node.parent.pos[1,0], node.pos[1,0]],
                  [node.parent.pos[2,0], node.pos[2,0]], color='g')
        node = node.parent

# This function counts points that lay between the starting and final position of a ball
def points_inside_trajectory(P, x1, x2, r, P_KD_tree=None):
    if P_KD_tree is None:
        P_KD_tree = spa.KDTree(P.T)

    # half of the x1 - x2 distance
    r2 = np.linalg.norm((x1 - x2)) * 0.5

    # points inside the big sphere
    P_inliers = kd_tree.query_ball_point(((x1 + x2) * 0.5).T, r2 + r)[0]
    P_inliers_filt = []

    # points inside the tube
    for i in P_inliers:
        if np.linalg.norm(np.cross((P[:, [i]] - x1).T, (P[:, [i]] - x2).T)) / np.linalg.norm(x2 - x1) < r:
            P_inliers_filt.append(i)

    P_inliers = list(P_inliers_filt)
    P_inliers_filt = []
    P_inliers_potential = []

    # points inside the smaller sphere
    small_radius = np.sqrt(r2 ** 2 + r ** 2)

    for i in P_inliers:
        if np.linalg.norm((x1 + x2) / 2 - P[:, [i]]) < small_radius:
            P_inliers_filt.append(i)
        else:
            P_inliers_potential.append(i)

    # points lying inside the smallest radii around x1 or x2
    for i in P_inliers_potential:
        if np.minimum(np.linalg.norm(P[:, [i]] - x1), np.linalg.norm(P[:, [i]] - x2)) < r:
            P_inliers_filt.append(i)

    # sort the inliers and return
    P_inliers_filt.sort()
    return P_inliers_filt

# Counts collision points after calling points inside trajectory and returns true of false
def isCollision(parent_node, pos, P_KD_Tree, P, R, COLLISION_THRESH=3):
    collisions = points_inside_trajectory(P, parent_node.pos, pos, R, P_KD_Tree)
    if len(collisions) > COLLISION_THRESH:
        return True
    else:
        return False

def samplePositionAtRandom():
    # # sample the space by a random point
    # # it is sligthly slower, because the samples often fall to the yard and the potential steps point into walls
    # rand = Node(np.random.random() * (XDIM[1]-XDIM[0]) + XDIM[0],
    #             np.random.random() * (YDIM[1]-YDIM[0]) + YDIM[0],
    #             np.random.random() * (ZDIM[1]-ZDIM[0]) + ZDIM[0])

    # slightly more informed sampling - only in the space of the building, not in the yard area
    # there are two areas, the A area is cca 95m2  and the B area 216m2, so 1/3 times goes A, the rest B
    # to get approximately uniform sampling in the given space
    if np.random.random() > 0.66:
        rand = Node(np.random.random() * (XDIM_A[1] - XDIM_A[0]) + XDIM_A[0],
                    np.random.random() * (YDIM_A[1] - YDIM_A[0]) + YDIM_A[0],
                    np.random.random() * (ZDIM_A[1] - ZDIM_A[0]) + ZDIM_A[0])
    else:
        rand = Node(np.random.random() * (XDIM_B[1] - XDIM_B[0]) + XDIM_B[0],
                    np.random.random() * (YDIM_B[1] - YDIM_B[0]) + YDIM_B[0],
                    np.random.random() * (ZDIM_B[1] - ZDIM_B[0]) + ZDIM_B[0])
    return rand

def plotSolution(P, nodes):
    fig = plt.figure()
    ax = fig.add_subplot(111, projection='3d')
    ax.axis('equal')
    plotNodes(nodes, ax)
    showPointClouds(P, ax)
    plt.draw()
    plt.pause(0.01)
    plt.show()



# main function
start = Node(2,-1,2)            # Start in the corner, in front of our office

goal = Node(21,0,1.5)           # easy goal node used to check if finished
#goal = Node(21,0,7)            # hard goal node used to check if finished

# Initialize search tree
nodes = []
nodes.append(start)

# While we have not reached max num of nodes
while len(nodes) <= NUMNODES:

    # sample position at random
    randPos = samplePositionAtRandom()

    # find the closest node to randPos
    nNode = ...

    # estimate new position (if ||randPos-nNode||<EPSILON, newPos=randPos otherwise shorten the step to EPSILON
    newPos = ...

    # if exist collision-free path from nNode to newPos then add newPos to the tree
    # syntax: isCollision(nNode, newPos, kd_tree, P, R)
    nodes.append( ... )

    # if ||newPos-goal||<GOAL_DISTANCE_THRESHOLD then terminate
    if ...

# plot solution into pointcloud
plotSolution(P, nodes)

